@@ -251,12 +251,12 @@ master_service_init(const char *name, enum master_service_flags flags,
 	   before we get to ioloop. the corresponding t_pop() is in
 	   master_service_init_finish(). */
 	if ((flags & MASTER_SERVICE_FLAG_NO_INIT_DATASTACK_FRAME) == 0)
-		datastack_frame_id = t_push(NULL);
+		datastack_frame_id = t_push("master_service_init");
 
 	/* ignore these signals as early as possible */
 	lib_signals_init();
-        lib_signals_ignore(SIGPIPE, TRUE);
-        lib_signals_ignore(SIGALRM, FALSE);
+	lib_signals_ignore(SIGPIPE, TRUE);
+	lib_signals_ignore(SIGALRM, FALSE);
 
 	if (getenv(MASTER_UID_ENV) == NULL)
 		flags |= MASTER_SERVICE_FLAG_STANDALONE;
@@ -304,16 +304,14 @@ master_service_init(const char *name, enum master_service_flags flags,
 	T_BEGIN {
 		master_service_init_socket_listeners(service);
 	} T_END;
-	service->want_ssl_settings = service->want_ssl_server ||
-		(service->flags & MASTER_SERVICE_FLAG_USE_SSL_SETTINGS) != 0;
 
 #ifdef HAVE_SSL
-	/* load SSL module if necessary */
-	if (service->want_ssl_settings) {
+	/* Load the SSL module if we already know it is necessary. It can also
+	   get loaded later on-demand. */
+	if (service->want_ssl_server) {
 		const char *error;
 		if (ssl_module_load(&error) < 0)
 			i_fatal("Cannot load SSL module: %s", error);
-		service->ssl_module_loaded = TRUE;
 	}
 #endif
 
@@ -661,6 +659,9 @@ void master_service_init_finish(struct master_service *service)
 	i_assert(!service->init_finished);
 	service->init_finished = TRUE;
 
+	/* From now on we'll abort() if exit() is called unexpectedly. */
+	lib_set_clean_exit(FALSE);
+
 	/* set default signal handlers */
 	if ((service->flags & MASTER_SERVICE_FLAG_STANDALONE) == 0)
 		sigint_flags |= LIBSIG_FLAG_RESTART;
@@ -724,15 +725,15 @@ static void master_service_import_environment_real(const char *import_environmen
 		if (value == NULL)
 			key = *envs;
 		else {
-			key = t_strdup_until(*envs, value);
-			env_put(*envs);
+			key = t_strdup_until(*envs, value++);
+			env_put(key, value);
 		}
 		array_push_back(&keys, &key);
 	}
 	array_append_zero(&keys);
 
 	value = t_strarray_join(array_front(&keys), " ");
-	env_put(t_strconcat(DOVECOT_PRESERVE_ENVS_ENV"=", value, NULL));
+	env_put(DOVECOT_PRESERVE_ENVS_ENV, value);
 }
 
 void master_service_import_environment(const char *import_environment)
@@ -1525,9 +1526,3 @@ bool version_string_verify_full(const char *line, const char *service_name,
 	} T_END;
 	return ret;
 }
-
-bool master_service_is_ssl_module_loaded(struct master_service *service)
-{
-	/* if this is TRUE, then ssl module is loaded by init */
-	return service->ssl_module_loaded;
-}
