@@ -22,6 +22,7 @@
 
 const char *const smtp_client_transaction_state_names[] = {
 	"new",
+	"pending",
 	"mail_from",
 	"rcpt_to",
 	"data",
@@ -74,10 +75,13 @@ static void
 smtp_client_transaction_mail_free(struct smtp_client_transaction_mail **_mail)
 {
 	struct smtp_client_transaction_mail *mail = *_mail;
-	struct smtp_client_transaction *trans = mail->trans;
 
+	if (mail == NULL)
+		return;
 	*_mail = NULL;
 
+	struct smtp_client_transaction *trans = mail->trans;
+
 	if (mail->cmd_mail_from != NULL)
 		smtp_client_command_abort(&mail->cmd_mail_from);
 	DLLIST2_REMOVE(&trans->mail_head, &trans->mail_tail, mail);
@@ -90,28 +94,37 @@ smtp_client_transaction_mail_replied(
 	const struct smtp_reply *reply)
 {
 	struct smtp_client_transaction_mail *mail = *_mail;
+
+	if (mail == NULL)
+		return;
+	*_mail = NULL;
+
 	smtp_client_command_callback_t *mail_callback = mail->mail_callback;
 	void *context = mail->context;
 
 	mail->mail_callback = NULL;
+	smtp_client_transaction_mail_free(&mail);
 
-	/* call the callback */
+	/* Call the callback */
 	if (mail_callback != NULL)
 		mail_callback(reply, context);
-
-	smtp_client_transaction_mail_free(_mail);
 }
 
 void smtp_client_transaction_mail_abort(
 	struct smtp_client_transaction_mail **_mail)
 {
 	struct smtp_client_transaction_mail *mail = *_mail;
+
+	if (mail == NULL)
+		return;
+	*_mail = NULL;
+
 	struct smtp_client_transaction *trans = mail->trans;
 
 	i_assert(trans->state <= SMTP_CLIENT_TRANSACTION_STATE_MAIL_FROM ||
 		 trans->state == SMTP_CLIENT_TRANSACTION_STATE_ABORTED);
 
-	smtp_client_transaction_mail_free(_mail);
+	smtp_client_transaction_mail_free(&mail);
 }
 
 static void
@@ -120,15 +133,19 @@ smtp_client_transaction_mail_fail_reply(
 	const struct smtp_reply *reply)
 {
 	struct smtp_client_transaction_mail *mail = *_mail;
+
+	if (mail == NULL)
+		return;
+	*_mail = NULL;
+
 	smtp_client_command_callback_t *callback = mail->mail_callback;
 	void *context = mail->context;
 
 	mail->mail_callback = NULL;
+	smtp_client_transaction_mail_free(&mail);
 
 	if (callback != NULL)
 		callback(reply, context);
-
-	smtp_client_transaction_mail_free(_mail);
 }
 
 /*
@@ -142,16 +159,17 @@ smtp_client_transaction_rcpt_update_event(
 	const char *to = smtp_address_encode(rcpt->rcpt_to);
 
 	event_set_append_log_prefix(rcpt->event,
-		t_strdup_printf("rcpt <%s>: ", str_sanitize(to, 128)));
+				    t_strdup_printf("rcpt <%s>: ",
+						    str_sanitize(to, 128)));
 	event_add_str(rcpt->event, "rcpt_to", to);
 	smtp_params_rcpt_add_to_event(&rcpt->rcpt_params, rcpt->event);
 }
 
 static struct smtp_client_transaction_rcpt *
-smtp_client_transaction_rcpt_new(
-	struct smtp_client_transaction *trans, pool_t pool,
-	const struct smtp_address *rcpt_to,
-	const struct smtp_params_rcpt *rcpt_params)
+smtp_client_transaction_rcpt_new(struct smtp_client_transaction *trans,
+				 pool_t pool,
+				 const struct smtp_address *rcpt_to,
+				 const struct smtp_params_rcpt *rcpt_params)
 {
 	struct smtp_client_transaction_rcpt *rcpt;
 
@@ -183,10 +201,14 @@ smtp_client_transaction_rcpt_free(
 	struct smtp_client_transaction_rcpt **_rcpt)
 {
 	struct smtp_client_transaction_rcpt *rcpt = *_rcpt;
-	struct smtp_client_transaction *trans = rcpt->trans;
 
+	if (rcpt == NULL)
+		return;
 	*_rcpt = NULL;
 
+	struct smtp_client_transaction *trans = rcpt->trans;
+
+	smtp_client_command_abort(&rcpt->cmd_rcpt_to);
 	if (trans->rcpts_send == rcpt)
 		trans->rcpts_send = rcpt->next;
 	if (trans->rcpts_data == rcpt)
@@ -207,8 +229,7 @@ smtp_client_transaction_rcpt_free(
 		trans->rcpts_aborted++;
 
 		smtp_reply_init(&failure,
-				SMTP_CLIENT_COMMAND_ERROR_ABORTED,
-				"Aborted");
+				SMTP_CLIENT_COMMAND_ERROR_ABORTED, "Aborted");
 		failure.enhanced_code = SMTP_REPLY_ENH_CODE(9, 0, 0);
 
 		struct event_passthrough *e =
@@ -231,6 +252,9 @@ smtp_client_transaction_rcpt_approved(
 	struct smtp_client_transaction_rcpt **_rcpt)
 {
 	struct smtp_client_transaction_rcpt *prcpt = *_rcpt;
+
+	i_assert(prcpt != NULL);
+
 	struct smtp_client_transaction *trans = prcpt->trans;
 	struct smtp_client_transaction_rcpt *rcpt;
 	pool_t pool;
@@ -238,7 +262,7 @@ smtp_client_transaction_rcpt_approved(
 	i_assert(prcpt->queued);
 
 	if (prcpt->external_pool) {
-		/* allocated externally; just remove it from the queue */
+		/* Allocated externally; just remove it from the queue */
 		prcpt->queued = FALSE;
 		if (trans->rcpts_send == prcpt)
 			trans->rcpts_send = prcpt->next;
@@ -248,7 +272,7 @@ smtp_client_transaction_rcpt_approved(
 
 		rcpt = prcpt;
 	} else {
-		/* move to transaction pool */
+		/* Move to transaction pool */
 		pool = trans->pool;
 		rcpt = p_new(pool, struct smtp_client_transaction_rcpt, 1);
 		rcpt->trans = trans;
@@ -261,11 +285,11 @@ smtp_client_transaction_rcpt_approved(
 		rcpt->event = prcpt->event;
 		event_ref(rcpt->event);
 
-		/* free the old object, thereby removing it from the queue */
+		/* Free the old object, thereby removing it from the queue */
 		smtp_client_transaction_rcpt_free(&prcpt);
 	}
 
-	/* recipient is approved */
+	/* Recipient is approved */
 	DLLIST2_APPEND(&trans->rcpts_head, &trans->rcpts_tail, rcpt);
 	trans->rcpts_count++;
 	if (trans->rcpts_data == NULL)
@@ -280,9 +304,11 @@ smtp_client_transaction_rcpt_denied(
 	const struct smtp_reply *reply)
 {
 	struct smtp_client_transaction_rcpt *prcpt = *_rcpt;
-	struct smtp_client_transaction *trans = prcpt->trans;
 
 	*_rcpt = NULL;
+	i_assert(prcpt != NULL);
+
+	struct smtp_client_transaction *trans = prcpt->trans;
 
 	trans->rcpts_denied++;
 	trans->rcpts_failed++;
@@ -293,7 +319,7 @@ smtp_client_transaction_rcpt_denied(
 	smtp_reply_add_to_event(reply, e);
 	e_debug(e->event(), "Denied");
 
-	/* not pending anymore */
+	/* Not pending anymore */
 	smtp_client_transaction_rcpt_free(&prcpt);
 }
 
@@ -303,6 +329,11 @@ smtp_client_transaction_rcpt_replied(
 	const struct smtp_reply *reply)
 {
 	struct smtp_client_transaction_rcpt *rcpt = *_rcpt;
+
+	*_rcpt = NULL;
+	if (rcpt == NULL)
+		return;
+
 	bool success = smtp_reply_is_success(reply);
 	smtp_client_command_callback_t *rcpt_callback = rcpt->rcpt_callback;
 	void *context = rcpt->context;
@@ -314,11 +345,11 @@ smtp_client_transaction_rcpt_replied(
 	rcpt->finished = !success;
 
 	if (success)
-		smtp_client_transaction_rcpt_approved(_rcpt);
+		smtp_client_transaction_rcpt_approved(&rcpt);
 	else
-		smtp_client_transaction_rcpt_denied(_rcpt, reply);
+		smtp_client_transaction_rcpt_denied(&rcpt, reply);
 
-	/* call the callback */
+	/* Call the callback */
 	if (rcpt_callback != NULL)
 		rcpt_callback(reply, context);
 }
@@ -327,6 +358,11 @@ void smtp_client_transaction_rcpt_abort(
 	struct smtp_client_transaction_rcpt **_rcpt)
 {
 	struct smtp_client_transaction_rcpt *rcpt = *_rcpt;
+
+	if (rcpt == NULL)
+		return;
+	*_rcpt = NULL;
+
 	struct smtp_client_transaction *trans = rcpt->trans;
 
 	i_assert(rcpt->queued || rcpt->external_pool);
@@ -334,7 +370,7 @@ void smtp_client_transaction_rcpt_abort(
 	i_assert(trans->state <= SMTP_CLIENT_TRANSACTION_STATE_RCPT_TO ||
 		 trans->state == SMTP_CLIENT_TRANSACTION_STATE_ABORTED);
 
-	smtp_client_transaction_rcpt_free(_rcpt);
+	smtp_client_transaction_rcpt_free(&rcpt);
 }
 
 static void
@@ -343,6 +379,11 @@ smtp_client_transaction_rcpt_fail_reply(
 	const struct smtp_reply *reply)
 {
 	struct smtp_client_transaction_rcpt *rcpt = *_rcpt;
+
+	if (rcpt == NULL)
+		return;
+	*_rcpt = NULL;
+
 	struct smtp_client_transaction *trans = rcpt->trans;
 	smtp_client_command_callback_t *callback;
 	void *context;
@@ -369,10 +410,10 @@ smtp_client_transaction_rcpt_fail_reply(
 	smtp_reply_add_to_event(reply, e);
 	e_debug(e->event(), "Failed");
 
+	smtp_client_transaction_rcpt_free(&rcpt);
+
 	if (callback != NULL)
 		callback(reply, context);
-
-	smtp_client_transaction_rcpt_free(_rcpt);
 }
 
 static void
@@ -426,8 +467,8 @@ smtp_client_transaction_result_event(struct smtp_client_transaction *trans,
 				     const struct smtp_reply *reply)
 {
 	struct event_passthrough *e;
-	unsigned int rcpts_aborted = trans->rcpts_aborted +
-				     trans->rcpts_queue_count;
+	unsigned int rcpts_aborted = (trans->rcpts_aborted +
+				      trans->rcpts_queue_count);
 
 	e = event_create_passthrough(trans->event)->
 		set_name("smtp_client_transaction_finished")->
@@ -478,10 +519,11 @@ smtp_client_transaction_create_empty(
 #undef smtp_client_transaction_create
 struct smtp_client_transaction *
 smtp_client_transaction_create(struct smtp_client_connection *conn,
-	const struct smtp_address *mail_from,
-	const struct smtp_params_mail *mail_params,
-	enum smtp_client_transaction_flags flags,
-	smtp_client_transaction_callback_t *callback, void *context)
+			       const struct smtp_address *mail_from,
+			       const struct smtp_params_mail *mail_params,
+			       enum smtp_client_transaction_flags flags,
+			       smtp_client_transaction_callback_t *callback,
+			       void *context)
 {
 	struct smtp_client_transaction *trans;
 
@@ -532,27 +574,23 @@ void smtp_client_transaction_abort(struct smtp_client_transaction *trans)
 
 	e_debug(trans->event, "Abort");
 
-	/* clean up */
+	/* Clean up */
 	i_stream_unref(&trans->data_input);
 	timeout_remove(&trans->to_send);
 	timeout_remove(&trans->to_finish);
 
 	trans->cmd_last = NULL;
 
-	/* abort any pending commands */
+	/* Abort any pending commands */
 	while (trans->mail_head != NULL) {
 		struct smtp_client_transaction_mail *mail = trans->mail_head;
 
-		if (mail->cmd_mail_from != NULL)
-			smtp_client_command_abort(&mail->cmd_mail_from);
 		smtp_client_transaction_mail_free(&mail);
 	}
 	while (trans->rcpts_queue_count > 0) {
 		struct smtp_client_transaction_rcpt *rcpt =
 			trans->rcpts_queue_head;
 
-		if (rcpt->cmd_rcpt_to != NULL)
-			smtp_client_command_abort(&rcpt->cmd_rcpt_to);
 		smtp_client_transaction_rcpt_free(&rcpt);
 	}
 	if (trans->cmd_data != NULL)
@@ -567,7 +605,7 @@ void smtp_client_transaction_abort(struct smtp_client_transaction *trans)
 
 	smtp_client_connection_abort_transaction(conn, trans);
 
-	/* abort if not finished */
+	/* Abort if not finished */
 	if (trans->state < SMTP_CLIENT_TRANSACTION_STATE_FINISHED) {
 		struct event_passthrough *e;
 
@@ -684,7 +722,7 @@ void smtp_client_transaction_destroy(struct smtp_client_transaction **_trans)
 }
 
 void smtp_client_transaction_fail_reply(struct smtp_client_transaction *trans,
-	const struct smtp_reply *reply)
+					const struct smtp_reply *reply)
 {
 	struct smtp_client_transaction_rcpt *rcpt, *rcpt_next;
 
@@ -696,7 +734,7 @@ void smtp_client_transaction_fail_reply(struct smtp_client_transaction *trans,
 
 	e_debug(trans->event, "Returning failure: %s", smtp_reply_log(reply));
 
-	/* hold a reference to prevent early destruction in a callback */
+	/* Hold a reference to prevent early destruction in a callback */
 	smtp_client_transaction_ref(trans);
 
 	trans->cmd_last = NULL;
@@ -707,8 +745,6 @@ void smtp_client_transaction_fail_reply(struct smtp_client_transaction *trans,
 	while (trans->mail_head != NULL) {
 		struct smtp_client_transaction_mail *mail = trans->mail_head;
 
-		if (mail->cmd_mail_from != NULL)
-			smtp_client_command_abort(&mail->cmd_mail_from);
 		smtp_client_transaction_mail_fail_reply(&mail, reply);
 	}
 
@@ -730,21 +766,21 @@ void smtp_client_transaction_fail_reply(struct smtp_client_transaction *trans,
 
 	/* DATA / RSET */
 	if (!trans->data_provided && !trans->reset) {
-		/* none of smtp_client_transaction_send() and
+		/* None of smtp_client_transaction_send() and
 		   smtp_client_transaction_reset() was called so far
 		 */
 	} else if (trans->cmd_data != NULL) {
-		/* the DATA command is still pending; handle the failure by
+		/* The DATA command is still pending; handle the failure by
 		   failing the DATA command. */
 		smtp_client_command_fail_reply(&trans->cmd_data, reply);
 	} else if (trans->cmd_rset != NULL) {
-		/* the RSET command is still pending; handle the failure by
+		/* The RSET command is still pending; handle the failure by
 		   failing the RSET command. */
 		smtp_client_command_fail_reply(&trans->cmd_rset, reply);
 	} else {
 		i_assert(!trans->reset);
 
-		/* the DATA command was not sent yet; call all DATA callbacks
+		/* The DATA command was not sent yet; call all DATA callbacks
 		   for the recipients that were previously accepted. */
 		rcpt = trans->rcpts_data;
 		while (rcpt != NULL) {
@@ -757,7 +793,7 @@ void smtp_client_transaction_fail_reply(struct smtp_client_transaction *trans,
 		trans->data_callback = NULL;
 	}
 
-	/* plug */
+	/* Plug */
 	if (trans->failure == NULL)
 		trans->failure = smtp_reply_clone(trans->pool, reply);
 	if (trans->cmd_plug != NULL)
@@ -767,18 +803,18 @@ void smtp_client_transaction_fail_reply(struct smtp_client_transaction *trans,
 	trans->failing = FALSE;
 
 	if (trans->data_provided || trans->reset) {
-		/* abort the transaction only if smtp_client_transaction_send()
+		/* Abort the transaction only if smtp_client_transaction_send()
 		   or  smtp_client_transaction_reset() was called (and if it is
 		   not aborted already) */
 		smtp_client_transaction_abort(trans);
 	}
 
-	/* drop reference held earlier in this function */
+	/* Drop reference held earlier in this function */
 	smtp_client_transaction_unref(&trans);
 }
 
 void smtp_client_transaction_fail(struct smtp_client_transaction *trans,
-	unsigned int status, const char *error)
+				  unsigned int status, const char *error)
 {
 	struct smtp_reply reply;
 
@@ -789,6 +825,7 @@ void smtp_client_transaction_fail(struct smtp_client_transaction *trans,
 void smtp_client_transaction_set_event(struct smtp_client_transaction *trans,
 				       struct event *event)
 {
+	i_assert(trans->conn != NULL);
 	event_unref(&trans->event);
 	trans->event = event_create(event);
 	event_set_forced_debug(trans->event, trans->conn->set.debug);
@@ -800,8 +837,8 @@ smtp_client_transaction_timeout(struct smtp_client_transaction *trans)
 {
 	struct smtp_reply reply;
 
-	smtp_reply_printf(&reply, 451,
-		"Remote server not answering "
+	smtp_reply_printf(
+		&reply, 451, "Remote server not answering "
 		"(transaction timed out while %s)",
 		smtp_client_transaction_get_state_destription(trans));
 	reply.enhanced_code = SMTP_REPLY_ENH_CODE(4, 4, 0);
@@ -810,17 +847,18 @@ smtp_client_transaction_timeout(struct smtp_client_transaction *trans)
 }
 
 void smtp_client_transaction_set_timeout(struct smtp_client_transaction *trans,
-	unsigned int timeout_msecs)
+					 unsigned int timeout_msecs)
 {
 	i_assert(trans->state < SMTP_CLIENT_TRANSACTION_STATE_FINISHED);
 
 	trans->finish_timeout_msecs = timeout_msecs;
 
 	if (trans->data_input != NULL && timeout_msecs > 0) {
-		/* adjust timeout if it is already started */
+		/* Adjust timeout if it is already started */
 		timeout_remove(&trans->to_finish);
 		trans->to_finish = timeout_add(trans->finish_timeout_msecs,
-			smtp_client_transaction_timeout, trans);
+					       smtp_client_transaction_timeout,
+					       trans);
 	}
 }
 
@@ -834,6 +872,7 @@ smtp_client_transaction_mail_cb(const struct smtp_reply *reply,
 	e_debug(trans->event, "Got MAIL reply: %s", smtp_reply_log(reply));
 
 	i_assert(mail != NULL);
+	i_assert(trans->conn != NULL);
 
 	if (success) {
 		if (trans->sender_accepted) {
@@ -846,7 +885,7 @@ smtp_client_transaction_mail_cb(const struct smtp_reply *reply,
 		trans->sender_accepted = TRUE;
 	}
 
-	/* plug command line pipeline if no RCPT commands are yet issued */
+	/* Plug command line pipeline if no RCPT commands are yet issued */
 	if (!trans->immediate && mail->next == NULL &&
 	    mail->cmd_mail_from == trans->cmd_last) {
 		trans->cmd_plug = trans->cmd_last =
@@ -885,8 +924,10 @@ smtp_client_transaction_mail_cb(const struct smtp_reply *reply,
 	if (!success && !trans->sender_accepted) {
 		if (trans->state > SMTP_CLIENT_TRANSACTION_STATE_MAIL_FROM)
 			smtp_client_transaction_fail_reply(trans, reply);
-		else if (trans->mail_failure == NULL)
-			trans->mail_failure = smtp_reply_clone(trans->pool, reply);
+		else if (trans->mail_failure == NULL) {
+			trans->mail_failure =
+				smtp_reply_clone(trans->pool, reply);
+		}
 	}
 }
 
@@ -916,8 +957,8 @@ smtp_client_transaction_add_mail(struct smtp_client_transaction *trans,
 	return mail;
 }
 
-static void smtp_client_transaction_connection_ready(
-	struct smtp_client_transaction *trans)
+static void
+smtp_client_transaction_connection_ready(struct smtp_client_transaction *trans)
 {
 	if (trans->state != SMTP_CLIENT_TRANSACTION_STATE_PENDING)
 		return;
@@ -938,6 +979,7 @@ void smtp_client_transaction_start(
 	struct smtp_client_transaction_mail *mail = trans->mail_head;
 
 	i_assert(trans->state == SMTP_CLIENT_TRANSACTION_STATE_NEW);
+	i_assert(trans->conn != NULL);
 
 	i_assert(mail != NULL);
 	event_add_str(trans->event, "mail_from",
@@ -983,7 +1025,7 @@ void smtp_client_transaction_start_empty(
 	i_assert(trans->mail_head == NULL);
 
 	(void)smtp_client_transaction_mail_new(trans, mail_from, mail_params);
-	
+
 	smtp_client_transaction_start(trans, mail_callback, context);
 }
 
@@ -993,9 +1035,11 @@ smtp_client_transaction_rcpt_cb(const struct smtp_reply *reply,
 {
 	struct smtp_client_transaction *trans = rcpt->trans;
 
+	i_assert(trans->conn != NULL);
+
 	e_debug(trans->event, "Got RCPT reply: %s", smtp_reply_log(reply));
 
-	/* plug command line pipeline if DATA command is not yet issued */
+	/* Plug command line pipeline if DATA command is not yet issued */
 	if (!trans->immediate && !trans->reset &&
 	    rcpt->cmd_rcpt_to == trans->cmd_last && trans->cmd_data == NULL) {
 		trans->cmd_plug = trans->cmd_last =
@@ -1040,7 +1084,7 @@ smtp_client_transaction_add_rcpt(struct smtp_client_transaction *trans,
 	i_assert(trans->state < SMTP_CLIENT_TRANSACTION_STATE_FINISHED);
 
 	if (trans->mail_head == NULL &&
-		trans->state == SMTP_CLIENT_TRANSACTION_STATE_MAIL_FROM)
+	    trans->state == SMTP_CLIENT_TRANSACTION_STATE_MAIL_FROM)
 		trans->state = SMTP_CLIENT_TRANSACTION_STATE_RCPT_TO;
 
 	pool = pool_alloconly_create("smtp transaction rcpt", 512);
@@ -1077,7 +1121,7 @@ smtp_client_transaction_add_pool_rcpt(
 	i_assert(trans->state < SMTP_CLIENT_TRANSACTION_STATE_FINISHED);
 
 	if (trans->mail_head == NULL &&
-		trans->state == SMTP_CLIENT_TRANSACTION_STATE_MAIL_FROM)
+	    trans->state == SMTP_CLIENT_TRANSACTION_STATE_MAIL_FROM)
 		trans->state = SMTP_CLIENT_TRANSACTION_STATE_RCPT_TO;
 
 	rcpt = smtp_client_transaction_rcpt_new(trans, pool,
@@ -1140,8 +1184,8 @@ smtp_client_transaction_data_cb(const struct smtp_reply *reply,
 
 	/* finished */
 	smtp_client_transaction_finish(
-		trans, (trans->data_failure == NULL ? reply :
-			trans->data_failure));
+		trans, (trans->data_failure == NULL ?
+			reply : trans->data_failure));
 
 	smtp_client_transaction_unref(&trans);
 }
@@ -1173,6 +1217,8 @@ smtp_client_transaction_send_data(struct smtp_client_transaction *trans)
 		}
 		i_assert(failure.status != 0);
 	} else {
+		i_assert(trans->conn != NULL);
+
 		trans->cmd_data = smtp_client_command_data_submit_after(
 			trans->conn, 0, trans->cmd_last, trans->data_input,
 			smtp_client_transaction_data_cb, trans);
@@ -1217,7 +1263,8 @@ void smtp_client_transaction_send(
 	if (trans->finish_timeout_msecs > 0) {
 		i_assert(trans->to_finish == NULL);
 		trans->to_finish = timeout_add(trans->finish_timeout_msecs,
-			smtp_client_transaction_timeout, trans);
+					       smtp_client_transaction_timeout,
+					       trans);
 	}
 
 	smtp_client_transaction_submit(trans, TRUE);
@@ -1235,7 +1282,7 @@ smtp_client_transaction_rset_cb(const struct smtp_reply *reply,
 		trans->reset_callback(reply, trans->reset_context);
 	trans->reset_callback = NULL;
 
-	/* finished */
+	/* Finished */
 	smtp_client_transaction_finish(trans, reply);
 
 	smtp_client_transaction_unref(&trans);
@@ -1244,24 +1291,38 @@ smtp_client_transaction_rset_cb(const struct smtp_reply *reply,
 static void
 smtp_client_transaction_send_reset(struct smtp_client_transaction *trans)
 {
+	struct smtp_reply failure;
+
 	i_assert(trans->reset);
 
 	e_debug(trans->event, "Sending reset");
 
 	timeout_remove(&trans->to_send);
 
-	trans->cmd_rset = smtp_client_command_rset_submit_after(
-		trans->conn, 0, trans->cmd_last,
-		smtp_client_transaction_rset_cb, trans);
+	i_zero(&failure);
+	if (trans->failure != NULL) {
+		smtp_client_transaction_fail_reply(trans, trans->failure);
+		failure = *trans->failure;
+		i_assert(failure.status != 0);
+	} else {
+		i_assert(trans->conn != NULL);
 
-	if (trans->cmd_last != NULL)
-		smtp_client_command_unlock(trans->cmd_last);
+		trans->cmd_rset = smtp_client_command_rset_submit_after(
+			trans->conn, 0, trans->cmd_last,
+			smtp_client_transaction_rset_cb, trans);
 
-	smtp_client_transaction_try_complete(trans);
+		if (trans->cmd_last != NULL)
+			smtp_client_command_unlock(trans->cmd_last);
+
+		smtp_client_transaction_try_complete(trans);
+	}
 
 	if (trans->cmd_plug != NULL)
 		smtp_client_command_abort(&trans->cmd_plug);
 	trans->cmd_last = NULL;
+
+	if (failure.status != 0)
+		smtp_client_transaction_finish(trans, &failure);
 }
 
 #undef smtp_client_transaction_reset
@@ -1283,7 +1344,8 @@ void smtp_client_transaction_reset(
 	if (trans->finish_timeout_msecs > 0) {
 		i_assert(trans->to_finish == NULL);
 		trans->to_finish = timeout_add(trans->finish_timeout_msecs,
-			smtp_client_transaction_timeout, trans);
+					       smtp_client_transaction_timeout,
+					       trans);
 	}
 
 	smtp_client_transaction_submit(trans, TRUE);
@@ -1306,6 +1368,8 @@ smtp_client_transaction_do_submit_more(struct smtp_client_transaction *trans)
 		return;
 	}
 
+	i_assert(trans->conn != NULL);
+
 	/* Make sure transaction is started */
 	if (trans->state == SMTP_CLIENT_TRANSACTION_STATE_NEW) {
 		enum smtp_client_transaction_state state;
@@ -1326,7 +1390,8 @@ smtp_client_transaction_do_submit_more(struct smtp_client_transaction *trans)
 	if (trans->mail_send != NULL) {
 		e_debug(trans->event, "Sending MAIL command");
 
-		i_assert(trans->state == SMTP_CLIENT_TRANSACTION_STATE_MAIL_FROM);
+		i_assert(trans->state ==
+			 SMTP_CLIENT_TRANSACTION_STATE_MAIL_FROM);
 
 		if (trans->cmd_last != NULL)
 			smtp_client_command_unlock(trans->cmd_last);
@@ -1337,7 +1402,8 @@ smtp_client_transaction_do_submit_more(struct smtp_client_transaction *trans)
 
 			trans->mail_send = trans->mail_send->next;
 			mail->cmd_mail_from = trans->cmd_last =
-				smtp_client_command_mail_submit(trans->conn, 0,
+				smtp_client_command_mail_submit(
+					trans->conn, 0,
 					mail->mail_from, &mail->mail_params,
 					smtp_client_transaction_mail_cb, trans);
 		}
@@ -1425,6 +1491,8 @@ smtp_client_transaction_submit(struct smtp_client_transaction *trans,
 static void
 smtp_client_transaction_try_complete(struct smtp_client_transaction *trans)
 {
+	i_assert(trans->conn != NULL);
+
 	if (trans->rcpts_queue_count > 0) {
 		/* Not all RCPT replies have come in yet */
 		e_debug(trans->event,  "RCPT replies are still pending (%u/%u)",
@@ -1502,6 +1570,13 @@ void smtp_client_transaction_connection_result(
 	smtp_client_transaction_connection_ready(trans);
 }
 
+void smtp_client_transaction_connection_destroyed(
+	struct smtp_client_transaction *trans)
+{
+	i_assert(trans->failure != NULL);
+	smtp_client_connection_unref(&trans->conn);
+}
+
 const struct smtp_client_transaction_times *
 smtp_client_transaction_get_times(struct smtp_client_transaction *trans)
 {
@@ -1518,7 +1593,7 @@ const char *
 smtp_client_transaction_get_state_name(struct smtp_client_transaction *trans)
 {
 	i_assert(trans->state >= SMTP_CLIENT_TRANSACTION_STATE_NEW &&
-		trans->state <= SMTP_CLIENT_TRANSACTION_STATE_ABORTED);
+		 trans->state <= SMTP_CLIENT_TRANSACTION_STATE_ABORTED);
 	return smtp_client_transaction_state_names[trans->state];
 }
 
@@ -1532,6 +1607,7 @@ smtp_client_transaction_get_state_destription(
 	case SMTP_CLIENT_TRANSACTION_STATE_NEW:
 		break;
 	case SMTP_CLIENT_TRANSACTION_STATE_PENDING:
+		i_assert(trans->conn != NULL);
 		conn_state = smtp_client_connection_get_state(trans->conn);
 		switch (conn_state) {
 		case SMTP_CLIENT_CONNECTION_STATE_CONNECTING:
