@@ -4,6 +4,8 @@
 #include "llist.h"
 #include "array.h"
 #include "str.h"
+#include "guid.h"
+#include "base64.h"
 #include "ioloop.h"
 #include "istream.h"
 #include "ostream.h"
@@ -282,7 +284,9 @@ static void smtp_server_connection_ready(struct smtp_server_connection *conn)
 				    smtp_server_connection_output, conn);
 
 	o_stream_cork(conn->conn.output);
-	if (conn->authenticated) {
+	if (conn->set.no_greeting) {
+		/* Don't send greeting or login reply. */
+	} else if (conn->authenticated) {
 		/* RFC 4954, Section 4:
 		   Should the client successfully complete the exchange, the
 		   SMTP server issues a 235 reply. */
@@ -622,7 +626,6 @@ static bool
 smtp_server_connection_next_reply(struct smtp_server_connection *conn)
 {
 	struct smtp_server_command *cmd;
-	unsigned int i;
 
 	cmd = conn->command_queue_head;
 	if (cmd == NULL) {
@@ -631,38 +634,7 @@ smtp_server_connection_next_reply(struct smtp_server_connection *conn)
 		return FALSE;
 	}
 
-	if (!smtp_server_command_next_to_reply(&cmd))
-		return FALSE;
-	if (cmd->state < SMTP_SERVER_COMMAND_STATE_READY_TO_REPLY)
-		return FALSE;
-
-	i_assert(cmd->state == SMTP_SERVER_COMMAND_STATE_READY_TO_REPLY &&
-		 array_is_created(&cmd->replies));
-
-	if (!smtp_server_command_completed(&cmd))
-		return TRUE;
-
-	/* Send command replies */
-	// FIXME: handle LMTP DATA command with enormous number of recipients;
-	// i.e. don't keep filling output stream with replies indefinitely.
-	for (i = 0; i < cmd->replies_expected; i++) {
-		struct smtp_server_reply *reply;
-
-		reply = array_idx_modifiable(&cmd->replies, i);
-
-		if (!reply->submitted) {
-			i_assert(!reply->sent);
-			cmd->state = SMTP_SERVER_COMMAND_STATE_PROCESSING;
-			break;
-		}
-		if (smtp_server_reply_send(reply) < 0)
-			return FALSE;
-	}
-	if (cmd->state == SMTP_SERVER_COMMAND_STATE_PROCESSING)
-		return FALSE;
-
-	smtp_server_command_finished(cmd);
-	return TRUE;
+	return smtp_server_command_send_replies(cmd);
 }
 
 void smtp_server_connection_cork(struct smtp_server_connection *conn)
@@ -781,6 +753,30 @@ smtp_server_connection_event_create(struct smtp_server *server,
 	return conn_event;
 }
 
+static void
+smtp_server_connection_update_event(struct smtp_server_connection *conn)
+{
+	event_add_str(conn->event, "connection_id", conn->session_id);
+	event_add_str(conn->event, "session", conn->session_id);
+}
+
+static void
+smtp_server_connection_init_session(struct smtp_server_connection *conn)
+{
+	guid_128_t guid;
+	string_t *session_id;
+
+	session_id = t_str_new(30);
+	guid_128_generate(guid);
+	base64_encode(guid, sizeof(guid), session_id);
+
+	/* drop trailing "==" */
+	i_assert(str_c(session_id)[str_len(session_id)-2] == '=');
+	str_truncate(session_id, str_len(session_id)-2);
+
+	conn->session_id = i_strdup(str_c(session_id));
+}
+
 static struct smtp_server_connection * ATTR_NULL(5, 6)
 smtp_server_connection_alloc(struct smtp_server *server,
 			     const struct smtp_server_settings *set,
@@ -886,6 +882,8 @@ smtp_server_connection_alloc(struct smtp_server *server,
 		conn->set.rcpt_domain_optional =
 			conn->set.rcpt_domain_optional ||
 				set->rcpt_domain_optional;
+		conn->set.no_greeting =
+			conn->set.no_greeting || set->no_greeting;
 		conn->set.debug = conn->set.debug || set->debug;
 	}
 
@@ -933,6 +931,8 @@ smtp_server_connection_alloc(struct smtp_server *server,
 			set->socket_recv_buffer_size);
 	}
 
+	smtp_server_connection_init_session(conn);
+
 	return conn;
 }
 
@@ -953,6 +953,7 @@ smtp_server_connection_create(
 	connection_init_server_ip(server->conn_list, &conn->conn, NULL,
 				  fd_in, fd_out, remote_ip, remote_port);
 	conn->event = conn->conn.event;
+	smtp_server_connection_update_event(conn);
 	event_unref(&conn_event);
 
 	conn->ssl_start = ssl_start;
@@ -996,6 +997,7 @@ smtp_server_connection_create_from_streams(
 				     input, output);
 	conn->created_from_streams = TRUE;
 	conn->event = conn->conn.event;
+	smtp_server_connection_update_event(conn);
 	event_unref(&conn_event);
 
 	/* Halt input until started */
@@ -1117,8 +1119,10 @@ bool smtp_server_connection_unref(struct smtp_server_connection **_conn)
 
 	connection_deinit(&conn->conn);
 
+	i_free(conn->proxy_helo);
 	i_free(conn->helo_domain);
 	i_free(conn->username);
+	i_free(conn->session_id);
 	event_unref(&conn->next_trans_event);
 	pool_unref(&conn->pool);
 	return FALSE;
@@ -1205,12 +1209,12 @@ void smtp_server_connection_login(struct smtp_server_connection *conn,
 				  unsigned int pdata_len, bool ssl_secured)
 {
 	i_assert(!conn->started);
-	i_assert(conn->username == NULL);
-	i_assert(conn->helo_domain == NULL);
 
 	conn->set.capabilities &= ENUM_NEGATE(SMTP_CAPABILITY_STARTTLS);
+	i_free(conn->username);
 	conn->username = i_strdup(username);
 	if (helo != NULL && *helo != '\0') {
+		i_free(conn->helo_domain);
 		conn->helo_domain = i_strdup(helo);
 		conn->helo.domain = conn->helo_domain;
 		conn->helo.domain_valid = TRUE;
@@ -1518,9 +1522,12 @@ void smtp_server_connection_get_proxy_data(struct smtp_server_connection *conn,
 	i_zero(proxy_data);
 	proxy_data->source_ip = conn->conn.remote_ip;
 	proxy_data->source_port = conn->conn.remote_port;
-	if (conn->helo.domain_valid)
+	if (conn->proxy_helo != NULL)
+		proxy_data->helo = conn->proxy_helo;
+	else if (conn->helo.domain_valid)
 		proxy_data->helo = conn->helo.domain;
 	proxy_data->login = conn->username;
+	proxy_data->session = conn->session_id;
 
 	if (conn->proxy_proto != SMTP_PROXY_PROTOCOL_UNKNOWN)
 		proxy_data->proto = conn->proxy_proto;
@@ -1548,6 +1555,10 @@ void smtp_server_connection_set_proxy_data(
 		conn->helo_domain = i_strdup(proxy_data->helo);
 		conn->helo.domain = conn->helo_domain;
 		conn->helo.domain_valid = TRUE;
+		if (conn->helo.domain_valid) {
+			i_free(conn->proxy_helo);
+			conn->proxy_helo = i_strdup(proxy_data->helo);
+		}
 	}
 	if (proxy_data->login != NULL) {
 		i_free(conn->username);
@@ -1555,24 +1566,27 @@ void smtp_server_connection_set_proxy_data(
 	}
 	if (proxy_data->proto != SMTP_PROXY_PROTOCOL_UNKNOWN)
 		conn->proxy_proto = proxy_data->proto;
+	if (proxy_data->session != NULL &&
+	    strcmp(proxy_data->session, conn->session_id) != 0) {
+		e_debug(conn->event, "Updated session ID from %s to %s",
+		        conn->session_id, proxy_data->session);
+		i_free(conn->session_id);
+		conn->session_id = i_strdup(proxy_data->session);
+	}
 
 	if (proxy_data->ttl_plus_1 > 0)
 		conn->proxy_ttl_plus_1 = proxy_data->ttl_plus_1;
 	if (conn->proxy_timeout_secs > 0)
 		conn->proxy_timeout_secs = proxy_data->timeout_secs;
 
+	connection_update_properties(&conn->conn);
+	smtp_server_connection_update_event(conn);
+
 	if (conn->callbacks != NULL &&
 	    conn->callbacks->conn_proxy_data_updated != NULL) {
 		struct smtp_proxy_data full_data;
 
-		i_zero(&full_data);
-		full_data.source_ip = conn->conn.remote_ip;
-		full_data.source_port = conn->conn.remote_port;
-		full_data.helo = conn->helo.domain;
-		full_data.login = conn->username;
-		full_data.proto = conn->proxy_proto;
-		full_data.ttl_plus_1 = conn->proxy_ttl_plus_1;
-		full_data.timeout_secs = conn->proxy_timeout_secs;
+		smtp_server_connection_get_proxy_data(conn, &full_data);
 
 		conn->callbacks->
 			conn_proxy_data_updated(conn->context, &full_data);
@@ -1621,3 +1635,14 @@ void smtp_server_connection_switch_ioloop(struct smtp_server_connection *conn)
 		conn->to_idle = io_loop_move_timeout(&conn->to_idle);
 	connection_switch_ioloop(&conn->conn);
 }
+
+struct event_reason *
+smtp_server_connection_reason_begin(struct smtp_server_connection *conn,
+				    const char *name)
+{
+	if (conn->set.reason_code_module == NULL)
+		return NULL;
+	const char *reason_code =
+		event_reason_code(conn->set.reason_code_module, name);
+	return event_reason_begin(reason_code);
+}
