@@ -57,13 +57,11 @@ imapc_mail_fetch_callback(const struct imapc_command_reply *reply,
 {
 	struct imapc_fetch_request *request = context;
 	struct imapc_fetch_request *const *requests;
-	struct imapc_mail *const *mailp;
+	struct imapc_mail *mail;
 	struct imapc_mailbox *mbox = NULL;
 	unsigned int i, count;
 
-	array_foreach(&request->mails, mailp) {
-		struct imapc_mail *mail = *mailp;
-
+	array_foreach_elem(&request->mails, mail) {
 		i_assert(mail->fetch_count > 0);
 		imapc_mail_set_failure(mail, reply);
 		if (--mail->fetch_count == 0)
@@ -215,11 +213,8 @@ imapc_mail_send_fetch(struct mail *_mail, enum mail_fetch_field fields,
 	i_assert(headers == NULL ||
 		 IMAPC_BOX_HAS_FEATURE(mbox, IMAPC_FEATURE_FETCH_HEADERS));
 
-	if (_mail->lookup_abort != MAIL_LOOKUP_ABORT_NEVER) {
-		mail_set_aborted(_mail);
+	if (!mail_stream_access_start(_mail))
 		return -1;
-	}
-	_mail->mail_stream_opened = TRUE;
 
 	/* drop any fields that we may already be fetching currently */
 	fields &= ENUM_NEGATE(mail->fetching_fields);
@@ -330,6 +325,9 @@ static void imapc_mail_cache_get(struct imapc_mail *mail,
 	}
 	mail->header_fetched = TRUE;
 	mail->body_fetched = TRUE;
+	/* The stream was already accessed and now it's cached.
+	   It still needs to be set accessed to avoid assert-crash. */
+	mail->imail.mail.mail.mail_stream_accessed = TRUE;
 	imapc_mail_init_stream(mail);
 }
 
@@ -512,15 +510,15 @@ int imapc_mail_fetch(struct mail *_mail, enum mail_fetch_field fields,
 void imapc_mail_fetch_flush(struct imapc_mailbox *mbox)
 {
 	struct imapc_command *cmd;
-	struct imapc_mail *const *mailp;
+	struct imapc_mail *mail;
 
 	if (mbox->pending_fetch_request == NULL) {
 		i_assert(mbox->to_pending_fetch_send == NULL);
 		return;
 	}
 
-	array_foreach(&mbox->pending_fetch_request->mails, mailp)
-		(*mailp)->fetch_sent = TRUE;
+	array_foreach_elem(&mbox->pending_fetch_request->mails, mail)
+		mail->fetch_sent = TRUE;
 
 	cmd = imapc_client_mailbox_cmd(mbox->client_box,
 				       imapc_mail_fetch_callback,
