@@ -81,7 +81,7 @@ int index_mail_cache_lookup_field(struct index_mail *mail, buffer_t *buf,
 	int ret;
 
 	ret = mail_cache_lookup_field(mail->mail.mail.transaction->cache_view,
-				      buf, mail->data.seq, field_idx);
+				      buf, mail->mail.mail.seq, field_idx);
 	if (ret > 0)
 		mail->mail.mail.transaction->stats.cache_hit_count++;
 
@@ -341,7 +341,7 @@ index_mail_get_keyword_indexes(struct mail *_mail)
 	if (!array_is_created(&data->keyword_indexes)) {
 		p_array_init(&data->keyword_indexes, mail->mail.data_pool, 32);
 		mail_index_lookup_keywords(_mail->transaction->view,
-					   mail->data.seq,
+					   mail->mail.mail.seq,
 					   &data->keyword_indexes);
 	}
 	return &data->keyword_indexes;
@@ -1195,8 +1195,15 @@ index_mail_parse_body_finish(struct index_mail *mail,
 		mail->data.parsed_bodystructure = FALSE;
 		if (mail->data.save_bodystructure_body)
 			mail->data.save_bodystructure_header = TRUE;
+		if (mail->data.header_parser_initialized)
+			index_mail_parse_header_deinit(mail);
 		return -1;
 	}
+	if (mail->data.header_parser_initialized) {
+		i_assert(!success);
+		index_mail_parse_header_deinit(mail);
+	}
+
 	if (mail->data.save_bodystructure_body) {
 		mail->data.parsed_bodystructure = TRUE;
 		mail->data.save_bodystructure_header = FALSE;
@@ -1340,7 +1347,7 @@ int index_mail_init_stream(struct index_mail *mail,
 	bool has_nuls, body_size_from_stream = FALSE;
 	int ret;
 
-	_mail->mail_stream_opened = TRUE;
+	i_assert(_mail->mail_stream_accessed);
 
 	if (!data->initialized_wrapper_stream &&
 	    _mail->transaction->stats_track) {
@@ -1361,15 +1368,21 @@ int index_mail_init_stream(struct index_mail *mail,
 
 	bool want_attachment_kw =
 		index_mail_want_attachment_keywords_on_fetch(mail);
-	if (want_attachment_kw)
+	if (want_attachment_kw) {
 		data->access_part |= PARSE_HDR | PARSE_BODY;
+		data->access_reason_code = "mail:attachment_keywords";
+	}
 
 	if (hdr_size != NULL || body_size != NULL)
 		(void)get_cached_msgpart_sizes(mail);
 
-	if (hdr_size != NULL || body_size != NULL || want_attachment_kw) {
+	bool want_body_parsing = want_attachment_kw ||
+		(body_size != NULL && !data->body_size_set &&
+		 (data->access_part & PARSE_BODY) != 0);
+
+	if (hdr_size != NULL || body_size != NULL || want_body_parsing) {
 		i_stream_seek(data->stream, 0);
-		if (!data->hdr_size_set || want_attachment_kw) {
+		if (!data->hdr_size_set || want_body_parsing) {
 			if ((data->access_part & (PARSE_HDR | PARSE_BODY)) != 0) {
 				(void)get_cached_parts(mail);
 				if (index_mail_parse_headers_internal(mail, NULL) < 0)
@@ -1389,10 +1402,10 @@ int index_mail_init_stream(struct index_mail *mail,
 			*hdr_size = data->hdr_size;
 	}
 
-	if (body_size != NULL || want_attachment_kw) {
+	if (body_size != NULL || want_body_parsing) {
 		if (!data->body_size_set && body_size != NULL)
 			index_mail_get_cached_body_size(mail);
-		if (!data->body_size_set || want_attachment_kw) {
+		if (!data->body_size_set || want_body_parsing) {
 			i_stream_seek(data->stream,
 				      data->hdr_size.physical_size);
 			if ((data->access_part & PARSE_BODY) != 0) {
@@ -1737,12 +1750,6 @@ index_mail_alloc(struct mailbox_transaction_context *t,
 	return &mail->mail.mail;
 }
 
-static void index_mail_init_event(struct mail *mail)
-{
-	mail->event = event_create(mail->box->event);
-	event_add_category(mail->event, &event_category_mail);
-}
-
 void index_mail_init(struct index_mail *mail,
 		     struct mailbox_transaction_context *t,
 		     enum mail_fetch_field wanted_fields,
@@ -1757,7 +1764,6 @@ void index_mail_init(struct index_mail *mail,
 	mail->mail.v = *t->box->mail_vfuncs;
 	mail->mail.mail.box = t->box;
 	mail->mail.mail.transaction = t;
-	index_mail_init_event(&mail->mail.mail);
 	t->mail_ref_count++;
 	if (data_pool != NULL)
 		mail->mail.data_pool = data_pool;
@@ -1849,7 +1855,7 @@ static void index_mail_reset_data(struct index_mail *mail)
 	mail->mail.mail.has_nuls = FALSE;
 	mail->mail.mail.has_no_nuls = FALSE;
 	mail->mail.mail.saving = FALSE;
-	mail->mail.mail.mail_stream_opened = FALSE;
+	mail->mail.mail.mail_stream_accessed = FALSE;
 	mail->mail.mail.mail_metadata_accessed = FALSE;
 }
 
@@ -1869,7 +1875,7 @@ void index_mail_close(struct mail *_mail)
 
 	/* make sure old mail isn't visible in the event anymore even if it's
 	   attempted to be used. */
-	event_unref(&_mail->event);
+	event_unref(&mail->mail._event);
 
 	/* If uid == 0 but seq != 0, we came here from saving a (non-mbox)
 	   message. If that happens, don't bother checking if anything should
@@ -1916,8 +1922,10 @@ static void check_envelope(struct index_mail *mail)
 						     "hdr.message-id");
 	if (cache_field_hdr == UINT_MAX ||
 	    mail_cache_field_exists(_mail->transaction->cache_view,
-				    _mail->seq, cache_field_hdr) <= 0)
+				    _mail->seq, cache_field_hdr) <= 0) {
+		mail->data.access_reason_code = "mail:imap_envelope";
 		mail->data.access_part |= PARSE_HDR;
+	}
 	mail->data.save_envelope = TRUE;
 }
 
@@ -1964,6 +1972,7 @@ void index_mail_update_access_parts_pre(struct mail *_mail)
 
 		if (mail_cache_field_exists(cache_view, _mail->seq,
 					    cache_field) <= 0) {
+			data->access_reason_code = "mail:mime_parts";
 			data->access_part |= PARSE_HDR | PARSE_BODY;
 			data->save_message_parts = TRUE;
 		}
@@ -1988,6 +1997,7 @@ void index_mail_update_access_parts_pre(struct mail *_mail)
 					    cache_field1) <= 0 &&
 		    mail_cache_field_exists(cache_view, _mail->seq,
 					    cache_field2) <= 0) {
+			data->access_reason_code = "mail:imap_bodystructure";
 			data->access_part |= PARSE_HDR | PARSE_BODY;
 			data->save_bodystructure_header = TRUE;
 			data->save_bodystructure_body = TRUE;
@@ -2003,6 +2013,7 @@ void index_mail_update_access_parts_pre(struct mail *_mail)
 
                 if (mail_cache_field_exists(cache_view, _mail->seq,
                                             cache_field) <= 0) {
+			data->access_reason_code = "mail:imap_bodystructure";
 			data->access_part |= PARSE_HDR | PARSE_BODY;
 			data->save_bodystructure_header = TRUE;
 			data->save_bodystructure_body = TRUE;
@@ -2017,6 +2028,7 @@ void index_mail_update_access_parts_pre(struct mail *_mail)
 
 		if (mail_cache_field_exists(cache_view, _mail->seq,
 					    cache_field) <= 0) {
+			data->access_reason_code = "mail:date";
 			data->access_part |= PARSE_HDR;
 			data->save_sent_date = TRUE;
 		}
@@ -2028,12 +2040,17 @@ void index_mail_update_access_parts_pre(struct mail *_mail)
 
 		if (mail_cache_field_exists(cache_view, _mail->seq,
 					    cache_field) <= 0) {
+			data->access_reason_code = "mail:snippet";
 			data->access_part |= PARSE_HDR | PARSE_BODY;
 			data->save_body_snippet = TRUE;
 		}
 	}
 	if ((data->wanted_fields & (MAIL_FETCH_STREAM_HEADER |
 				    MAIL_FETCH_STREAM_BODY)) != 0) {
+		/* Clear reason_code if set. The mail is going to be read
+		   in any case, so the previous reason for deciding to open
+		   the mail won't matter. */
+		data->access_reason_code = NULL;
 		if ((data->wanted_fields & MAIL_FETCH_STREAM_HEADER) != 0)
 			data->access_part |= READ_HDR;
 		if ((data->wanted_fields & MAIL_FETCH_STREAM_BODY) != 0)
@@ -2086,38 +2103,14 @@ void index_mail_update_access_parts_post(struct mail *_mail)
 	}
 }
 
-static bool index_mail_get_age_days(struct mail *mail, int *days_r)
-{
-	int age_days;
-	const struct mail_index_header *hdr =
-		mail_index_get_header(mail->transaction->view);
-	int n_days = N_ELEMENTS(hdr->day_first_uid);
-
-	for (age_days = 0; age_days < n_days; age_days++) {
-		if (mail->uid >= hdr->day_first_uid[age_days])
-			break;
-	}
-
-	if (age_days == n_days) {
-		/* mail is too old, cannot determine its age from
-		   day_first_uid[]. */
-		return FALSE;
-	}
-
-	if (hdr->day_stamp != 0) {
-		/* offset for hdr->day_stamp */
-		age_days += (ioloop_time - hdr->day_stamp) / (3600 * 24);
-	}
-	*days_r = age_days;
-	return TRUE;
-}
-
 void index_mail_set_seq(struct mail *_mail, uint32_t seq, bool saving)
 {
 	struct index_mail *mail = INDEX_MAIL(_mail);
-	int age_days;
+	const struct mail_index_record *rec;
+	struct mail_index_map *map;
+	bool expunged;
 
-	if (mail->data.seq == seq) {
+	if (mail->mail.mail.seq == seq) {
 		if (!saving)
 			return;
 		/* we started saving a mail, aborted it, and now we're saving
@@ -2127,24 +2120,17 @@ void index_mail_set_seq(struct mail *_mail, uint32_t seq, bool saving)
 
 	mail->mail.v.close(&mail->mail.mail);
 
-	mail->data.seq = seq;
 	mail->mail.mail.seq = seq;
 	mail->mail.mail.saving = saving;
-	mail_index_lookup_uid(_mail->transaction->view, seq,
-			      &mail->mail.mail.uid);
+
+	rec = mail_index_lookup_full(_mail->transaction->view, seq,
+				     &map, &expunged);
+	mail->mail.mail.uid = rec->uid;
 
 	/* Recreate the mail event when changing mails. Even though the same
-	   mail struct is reused, they are practically different mails. */
-	event_unref(&_mail->event);
-	index_mail_init_event(_mail);
-	event_add_int(_mail->event, "seq", _mail->seq);
-	event_add_int(_mail->event, "uid", _mail->uid);
-	/* Add mail age field to event. */
-	if (index_mail_get_age_days(_mail, &age_days))
-		event_add_int(_mail->event, "mail_age_days", age_days);
-
-	event_set_append_log_prefix(_mail->event, t_strdup_printf(
-		"%sUID %u: ", saving ? "saving " : "", _mail->uid));
+	   mail struct is reused, they are practically different mails. The
+	   event should have already been freed by close(). */
+	i_assert(mail->mail._event == NULL);
 
 	if (mail_index_view_is_inconsistent(_mail->transaction->view)) {
 		mail_set_expunged(&mail->mail.mail);
@@ -2153,7 +2139,7 @@ void index_mail_set_seq(struct mail *_mail, uint32_t seq, bool saving)
 	/* Allow callers to easily find out if this mail was already expunged
 	   by another session. It's possible that it could still be
 	   successfully accessed. */
-	if (mail_index_is_expunged(_mail->transaction->view, seq))
+	if (expunged)
 		mail_set_expunged(&mail->mail.mail);
 
 	if (!mail->mail.search_mail) {
@@ -2286,7 +2272,7 @@ void index_mail_free(struct mail *_mail)
 
 	mailbox_header_lookup_unref(&mail->data.wanted_headers);
 	mailbox_header_lookup_unref(&mail->mail.wanted_headers);
-	event_unref(&_mail->event);
+	event_unref(&mail->mail._event);
 	pool_unref(&mail->mail.data_pool);
 	pool_unref(&mail->mail.pool);
 }
@@ -2325,6 +2311,7 @@ void index_mail_cache_parse_deinit(struct mail *_mail, time_t received_date,
 
 		if (mail->data.parser_ctx == NULL) {
 			/* we didn't even start cache parsing */
+			i_assert(!mail->data.header_parser_initialized);
 			return;
 		}
 	}
@@ -2588,7 +2575,14 @@ void index_mail_set_cache_corrupted(struct mail *mail,
 int index_mail_opened(struct mail *mail,
 		      struct istream **stream ATTR_UNUSED)
 {
+	struct index_mail *imail =
+		container_of(mail, struct index_mail, mail.mail);
+	struct event_reason *reason = NULL;
+
+	if (imail->data.access_reason_code != NULL)
+		reason = event_reason_begin(imail->data.access_reason_code);
 	mail_opened_event(mail);
+	event_reason_end(&reason);
 	return 0;
 }
 
