@@ -123,7 +123,7 @@ index_list_get_path(struct mailbox_list *_list, const char *name,
 	struct mailbox_list_index_node *node;
 	struct mailbox_status status;
 	guid_128_t mailbox_guid;
-	const char *root_dir;
+	const char *root_dir, *reason;
 	uint32_t seq;
 	int ret;
 
@@ -189,7 +189,8 @@ index_list_get_path(struct mailbox_list *_list, const char *name,
 				       T_MAIL_ERR_MAILBOX_NOT_FOUND(name));
 		ret = -1;
 	} else if (!mailbox_list_index_status(_list, view, seq, 0,
-					      &status, mailbox_guid, NULL) ||
+					      &status, mailbox_guid,
+					      NULL, &reason) ||
 		   guid_128_is_empty(mailbox_guid)) {
 		mailbox_list_set_error(_list, MAIL_ERROR_NOTFOUND,
 				       T_MAIL_ERR_MAILBOX_NOT_FOUND(name));
@@ -454,7 +455,7 @@ index_list_mailbox_update(struct mailbox *box,
 	/* rename the directory */
 	if (!guid_128_is_empty(update->mailbox_guid) && old_path != NULL &&
 	    mailbox_list_set_get_root_path(&box->list->set,
-					   MAILBOX_LIST_PATH_TYPE_MAILBOX,
+					   MAILBOX_LIST_PATH_TYPE_DIR,
 					   &root_dir)) {
 		new_path = index_get_guid_path(box->list, root_dir,
 					       update->mailbox_guid);
