@@ -403,13 +403,7 @@ struct client *client_create(int fd_in, int fd_out,
 	o_stream_set_no_error_handling(client->output, TRUE);
 	o_stream_set_flush_callback(client->output, client_output, client);
 
-	if (set->rawlog_dir[0] != '\0') {
-		(void)iostream_rawlog_create(set->rawlog_dir, &client->input,
-					     &client->output);
-	}
-
 	p_array_init(&client->module_contexts, client->pool, 5);
-	client->io = io_add_istream(client->input, client_input, client);
         client->last_input = ioloop_time;
 	client->to_idle = timeout_add(CLIENT_IDLE_TIMEOUT_MSECS,
 				      client_idle_timeout, client);
@@ -442,6 +436,15 @@ struct client *client_create(int fd_in, int fd_out,
 	return client;
 }
 
+void client_create_finish(struct client *client)
+{
+	if (client->set->rawlog_dir[0] != '\0') {
+		(void)iostream_rawlog_create(client->set->rawlog_dir,
+					     &client->input, &client->output);
+	}
+	client->io = io_add_istream(client->input, client_input, client);
+}
+
 int client_init_mailbox(struct client *client, const char **error_r)
 {
         enum mailbox_flags flags;
@@ -454,7 +457,6 @@ int client_init_mailbox(struct client *client, const char **error_r)
 	if (!client->set->pop3_no_flag_updates)
 		flags |= MAILBOX_FLAG_DROP_RECENT;
 	client->mailbox = mailbox_alloc(client->inbox_ns->list, "INBOX", flags);
-	mailbox_set_reason(client->mailbox, "POP3 INBOX");
 	if (mailbox_open(client->mailbox) < 0) {
 		*error_r = t_strdup_printf("Couldn't open INBOX: %s",
 			mailbox_get_last_internal_error(client->mailbox, NULL));
@@ -742,8 +744,13 @@ bool client_handle_input(struct client *client)
 			*args++ = '\0';
 
 		T_BEGIN {
+			const char *reason_code =
+				event_reason_code_prefix("pop3", "cmd_", line);
+			struct event_reason *reason =
+				event_reason_begin(reason_code);
 			ret = client_command_execute(client, line,
 						     args != NULL ? args : "");
+			event_reason_end(&reason);
 		} T_END;
 		if (ret >= 0) {
 			client->bad_counter = 0;
