@@ -61,6 +61,7 @@ static const struct setting_define submission_login_setting_defines[] = {
 	DEF(STR, hostname),
 
 	DEF(SIZE, submission_max_mail_size),
+	DEF(STR, submission_client_workarounds),
 	DEF(STR, submission_backend_capabilities),
 
 	SETTING_DEFINE_LIST_END
@@ -70,6 +71,7 @@ static const struct submission_login_settings submission_login_default_settings
 	.hostname = "",
 
 	.submission_max_mail_size = 0,
+	.submission_client_workarounds = "",
 	.submission_backend_capabilities = NULL
 };
 
@@ -87,9 +89,7 @@ const struct setting_parser_info submission_login_setting_parser_info = {
 	.struct_size = sizeof(struct submission_login_settings),
 	.parent_offset = SIZE_MAX,
 
-#ifndef CONFIG_BINARY
 	.check_func = submission_login_settings_check,
-#endif
 	.dependencies = submission_login_setting_dependencies
 };
 
@@ -99,13 +99,66 @@ const struct setting_parser_info *submission_login_setting_roots[] = {
 	NULL
 };
 
+/* <settings checks> */
+struct submission_login_client_workaround_list {
+	const char *name;
+	enum submission_login_client_workarounds num;
+};
+
+/* These definitions need to be kept in sync with equivalent definitions present
+   in src/submission/submission-settings.c. Workarounds that are not relevant
+   to the submission-login service are defined as 0 here to prevent "Unknown
+   workaround" errors below. */
+static const struct submission_login_client_workaround_list
+submission_login_client_workaround_list[] = {
+	{ "whitespace-before-path", 0},
+	{ "mailbox-for-path", 0 },
+	{ "implicit-auth-external",
+	  SUBMISSION_LOGIN_WORKAROUND_IMPLICIT_AUTH_EXTERNAL },
+	{ NULL, 0 }
+};
+
+static int
+submission_login_settings_parse_workarounds(
+	struct submission_login_settings *set, const char **error_r)
+{
+	enum submission_login_client_workarounds client_workarounds = 0;
+        const struct submission_login_client_workaround_list *list;
+	const char *const *str;
+
+        str = t_strsplit_spaces(set->submission_client_workarounds, " ,");
+	for (; *str != NULL; str++) {
+		list = submission_login_client_workaround_list;
+		for (; list->name != NULL; list++) {
+			if (strcasecmp(*str, list->name) == 0) {
+				client_workarounds |= list->num;
+				break;
+			}
+		}
+		if (list->name == NULL) {
+			*error_r = t_strdup_printf(
+				"submission_client_workarounds: "
+				"Unknown workaround: %s", *str);
+			return -1;
+		}
+	}
+	set->parsed_workarounds = client_workarounds;
+	return 0;
+}
+
 static bool
-submission_login_settings_check(void *_set, pool_t pool,
-			const char **error_r ATTR_UNUSED)
+submission_login_settings_check(void *_set, pool_t pool ATTR_UNUSED,
+				const char **error_r)
 {
 	struct submission_login_settings *set = _set;
 
+	if (submission_login_settings_parse_workarounds(set, error_r) < 0)
+		return FALSE;
+
+#ifndef CONFIG_BINARY
 	if (*set->hostname == '\0')
 		set->hostname = p_strdup(pool, my_hostdomain());
+#endif
 	return TRUE;
 }
+/* </settings checks> */
