@@ -47,6 +47,8 @@ backend_relay_handle_relay_reply(struct submission_backend_relay *backend,
 				 const struct smtp_reply *reply,
 				 struct smtp_reply *reply_r) ATTR_NULL(2)
 {
+	struct client *client = backend->backend.client;
+	struct mail_user *user = client->user;
 	const char *enh_code, *msg, *log_msg = NULL;
 	const char *const *reply_lines;
 	bool result = TRUE;
@@ -126,8 +128,15 @@ backend_relay_handle_relay_reply(struct submission_backend_relay *backend,
 
 		reason = t_strdup_printf("%s%s", msg, detail);
 		smtp_client_transaction_destroy(&backend->trans);
-		if (log_msg != NULL)
-			i_error("%s: %s", log_msg, smtp_reply_log(reply));
+		if (log_msg != NULL) {
+			if (smtp_reply_is_remote(reply)) {
+				i_error("%s: %s",
+					log_msg, smtp_reply_log(reply));
+			} else if (user->mail_debug) {
+				i_debug("%s: %s",
+					log_msg, smtp_reply_log(reply));
+			}
+		}
 		submission_backend_fail(&backend->backend, cmd,
 					enh_code, reason);
 		return FALSE;
@@ -1057,7 +1066,6 @@ submission_backend_relay_create(
 	submission_backend_init(&backend->backend, pool, client,
 				&backend_relay_vfuncs);
 
-	i_zero(&ssl_set);
 	mail_user_init_ssl_client_settings(user, &ssl_set);
 	if (set->ssl_verify)
 		ssl_set.verbose_invalid_cert = TRUE;
@@ -1077,14 +1085,11 @@ submission_backend_relay_create(
 	}
 
 	if (set->trusted) {
-		struct smtp_server_helo_data *helo_data =
-			smtp_server_connection_get_helo_data(client->conn);
-
 		backend->trusted = TRUE;
 		smtp_set.peer_trusted = TRUE;
 
-		smtp_set.proxy_data.helo = helo_data->domain;
-		smtp_set.proxy_data.proto = SMTP_PROXY_PROTOCOL_ESMTP;
+		smtp_server_connection_get_proxy_data(client->conn,
+						      &smtp_set.proxy_data);
 
 		if (user->conn.remote_ip != NULL) {
 			smtp_set.proxy_data.source_ip =
